@@ -20,8 +20,9 @@
 
 package Lintian::fonts::opentype;
 
-use strict;
+use v5.20;
 use warnings;
+use utf8;
 use autodie qw(open);
 
 use Font::TTF::Font;
@@ -36,7 +37,7 @@ use namespace::clean;
 
 with 'Lintian::Check';
 
-sub files {
+sub visit_installed_files {
     my ($self, $file) = @_;
 
     return
